import 'dart:async';
import 'dart:convert';

//import 'package:cached_network_image/cached_network_image.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:google_fonts/google_fonts.dart';
import 'package:image_picker/image_picker.dart';
import 'package:intl/intl.dart';
import 'package:flutter_bloc/flutter_bloc.dart';

import '../ChatConfig.dart';
import '../const.dart';
import '../fullPhoto.dart';
import '../imagepicker/shared.dart';
import 'bloc/chat_bloc.dart';
import 'bloc/chat_event.dart';
import 'bloc/chat_state.dart';
import 'data/chat_repository.dart';
import 'data/user_repository.dart';
import 'model/message.dart';

class ChatScreen extends StatefulWidget {
  final String chatId;
  final String peerAvatar;
  final String userId;

  final UserRepository userRepository;
  final ChatRepository chatRepository;
  final ChatConfig chatConfig;

  final Function(BuildContext context) onDetailsClick;

  ChatScreen(
      {Key key,
      @required this.chatId,
      @required this.peerAvatar,
      @required this.userId,
      @required this.userRepository,
      @required this.chatRepository,
      @required this.chatConfig,
      this.onDetailsClick})
      : super(key: key);

  @override
  State createState() =>
      new ChatScreenState(chatId: chatId, peerAvatar: peerAvatar);
}

class ChatScreenState extends State<ChatScreen> {
  ChatScreenState(
      {Key key,
      @required this.chatId,
      @required this.peerAvatar});

  String chatId;
  String peerAvatar;

  final TextEditingController textEditingController =
      new TextEditingController();
  final ScrollController listScrollController = new ScrollController();
  final FocusNode focusNode = new FocusNode();

  ChatBloc _chatBloc;

  @override
  void initState() {
    super.initState();
    focusNode.addListener(onFocusChange);

    _chatBloc = ChatBloc(
        chatId, widget.chatRepository, widget.userRepository, widget.chatConfig)
      ..add(ChatOpened());

    _initImages();
  }

  Future _initImages() async {
    // >> To get paths you need these 2 lines
    final manifestContent =
        await DefaultAssetBundle.of(context).loadString('AssetManifest.json');

    final Map<String, dynamic> manifestMap = json.decode(manifestContent);

    print('assets $manifestMap');
  }

  void onFocusChange() {
    if (focusNode.hasFocus) {}
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      backgroundColor: Colors.white,
      body: BlocProvider<ChatBloc>(
        create: (BuildContext context) {
          return _chatBloc;
        },
        child: Stack(
          children: <Widget>[
            Column(
              children: <Widget>[
                buildTopBar(),
                buildMessagesSection(),
                // Input content
                buildInput(context),
              ],
            ),
            // Loading
          ],
        ),
      ),
    );
  }

  Future getImage(BuildContext context) async {
    try {
      final imageFile = await pickImage();

      if (imageFile != null) {
        BlocProvider.of<ChatBloc>(context).add(SendImage(imageFile));
      } else {
        print('imageFile == null');
        final lostData = await ImagePicker.retrieveLostData();
        print('lostData?.file==null ${lostData?.file == null}');
      }
    } catch (e) {
      print('getImage exception ${e}');
    }
  }

  void onSendMessage(BuildContext context, String content, int type) {
    // type: 0 = text, 1 = image
    if (content.trim() != '') {
      textEditingController.clear();

      var message = Message(
          idFrom: widget.userId,
          time: DateTime.now(),
          content: content,
          type: type,
          isLast: true,
          isMy: true);
      var sendMessageEvent = SendMessageEvent(message);
      BlocProvider.of<ChatBloc>(context).add(sendMessageEvent);

      if (kIsWeb) {
        focusNode.requestFocus();
      }
      listScrollController.animateTo(0.0,
          duration: Duration(milliseconds: 300), curve: Curves.easeOut);
    } else {
//      Fluttertoast.showToast(msg: 'Nothing to send');
    }
  }

  Widget buildItem(int index, Message message) {
    final messageWidget =
        message.isMy ? buildMyItem(message) : buildPeerItem(message);

    String dateText = message.headerText(_messages.safeGet(index + 1));

    return (dateText == null)
        ? messageWidget
        : wrapWithDateHeader(messageWidget, dateText);
  }

  Widget wrapWithDateHeader(Widget messageWidget, String dateText) {
    return Column(children: <Widget>[
      Row(
        mainAxisSize: MainAxisSize.max,
        children: <Widget>[
          Flexible(
              child: Divider(
            height: 1,
            color: Color(0xffEDF1F9),
          )),
          Container(
            padding: EdgeInsets.symmetric(vertical: 2, horizontal: 18),
            margin: EdgeInsets.symmetric(vertical: 15),
            child: Text(
              dateText,
              style: GoogleFonts.montserrat(
                  textStyle: TextStyle(fontSize: 12, color: Color(0xff55658A))),
            ),
            decoration: BoxDecoration(
                borderRadius: BorderRadius.circular(16),
                border: Border.all(
                  color: Color(0xffEDF1F9),
                  width: 1,
                )),
          ),
          Flexible(
              child: Divider(
            height: 1,
            color: Color(0xffEDF1F9),
          )),
        ],
      ),
      messageWidget
    ]);
  }

  Container buildPeerItem(Message message) {
    return Container(
      padding: const EdgeInsets.only(top: 8.0, bottom: 8.0),
      child: Column(
        children: <Widget>[
          Row(
            crossAxisAlignment: CrossAxisAlignment.start,
            children: <Widget>[
              !message.isMy && peerAvatar != null
                  ? Material(
                      child: Image(
                        image: NetworkImage(peerAvatar),
                        width: 22.0,
                        height: 2.0,
                        fit: BoxFit.cover,
                      ),
                      borderRadius: BorderRadius.all(
                        Radius.circular(18.0),
                      ),
                      clipBehavior: Clip.hardEdge,
                    )
                  : kIsWeb
                      ? Image.asset(
                          'packages/${widget.chatConfig.isMobileLibrary ? 'supchat/chat_module/' : 'supchat_chat/'}assets/icons/user.png',
                        )
                      : Icon(
                          Icons.face,
                          size: 22.0,
                          color: greyColor,
                        ),
              message.type == 0
                  ? Container(
                      constraints: BoxConstraints(maxWidth: 240),
                      child: Row(
                        crossAxisAlignment: CrossAxisAlignment.start,
                        mainAxisSize: MainAxisSize.min,
                        children: <Widget>[
                          Flexible(
                            child: Text(message.content,
                                style: GoogleFonts.montserrat(
                                    textStyle: TextStyle(
                                        color: Color(0xff55658A),
                                        fontSize: 12))),
                          ),
                          SizedBox(
                            width: 24,
                          ),
                          Text(
                            message.time.formatTime(),
                            overflow: TextOverflow.clip,
                            style: GoogleFonts.montserrat(
                                textStyle: TextStyle(
                                    color: Color(0xff55658A), fontSize: 10)),
                          ),
                        ],
                      ),
                      padding: EdgeInsets.fromLTRB(15.0, 10.0, 15.0, 10.0),
                      decoration: BoxDecoration(
                          color: Color(0xffF5F5F7),
                          borderRadius: BorderRadius.only(
                              topLeft: Radius.circular(5),
                              topRight: Radius.circular(15),
                              bottomLeft: Radius.circular(15),
                              bottomRight: Radius.circular(15))),
                      margin: EdgeInsets.only(left: 16.0),
                    )
                  : Container(
                      child: FlatButton(
                        child: Material(
                          child: Image(
                            image: NetworkImage(message.content),
                            width: 200.0,
                            height: 200.0,
                            fit: BoxFit.cover,
                          ),
                          borderRadius: BorderRadius.only(
                              topLeft: Radius.circular(5),
                              topRight: Radius.circular(15),
                              bottomLeft: Radius.circular(15),
                              bottomRight: Radius.circular(15)),
                          clipBehavior: Clip.hardEdge,
                        ),
                        onPressed: () {
                          Navigator.push(
                              context,
                              MaterialPageRoute(
                                  builder: (context) =>
                                      FullPhoto(url: message.content)));
                        },
                        padding: EdgeInsets.all(0),
                      ),
                      margin: EdgeInsets.only(left: 10.0),
                    )
            ],
          ),
        ],
        crossAxisAlignment: CrossAxisAlignment.start,
      ),
    );
  }

  Row buildMyItem(Message message) {
    return Row(
      children: <Widget>[
        Padding(
          padding: const EdgeInsets.only(top: 8.0, bottom: 8.0),
          child: Center(
            child: message.type == 0
                ? Container(
                    constraints: BoxConstraints(maxWidth: 240),
                    child: Row(
                      mainAxisSize: MainAxisSize.min,
                      crossAxisAlignment: CrossAxisAlignment.start,
                      children: <Widget>[
                        Flexible(
                          child: Text(
                            message.content,
                            style: GoogleFonts.montserrat(
                                textStyle: TextStyle(
                                    color: Colors.white, fontSize: 12)),
                          ),
                        ),
                        SizedBox(
                          width: 24,
                        ),
                        Text(
                          message.time.formatTime(),
                          style: GoogleFonts.montserrat(
                              textStyle:
                                  TextStyle(color: Colors.white, fontSize: 10)),
                        ),
                      ],
                    ),
                    padding: EdgeInsets.fromLTRB(15.0, 10.0, 15.0, 10.0),
//                    width: 200.0,
                    decoration: BoxDecoration(
                        color: Color(0xff6CA0FF),
                        borderRadius: BorderRadius.only(
                            topRight: Radius.circular(5),
                            topLeft: Radius.circular(15),
                            bottomLeft: Radius.circular(15),
                            bottomRight: Radius.circular(15))),
                  )
                : Container(
                    child: FlatButton(
                      child: Material(
                        child: Image(
                          image: NetworkImage(message.content),
                          width: 200.0,
                          height: 200.0,
                          fit: BoxFit.cover,
                        ),
                        borderRadius: BorderRadius.only(
                            topRight: Radius.circular(5),
                            topLeft: Radius.circular(15),
                            bottomLeft: Radius.circular(15),
                            bottomRight: Radius.circular(15)),
                        clipBehavior: Clip.hardEdge,
                      ),
                      onPressed: () {
                        Navigator.push(
                            context,
                            MaterialPageRoute(
                                builder: (context) =>
                                    FullPhoto(url: message.content)));
                      },
                      padding: EdgeInsets.all(0),
                    ),
                    margin: EdgeInsets.only(
                        bottom: message.isLast && message.isMy ? 20.0 : 10.0,
                        right: 10.0),
                  ),
          ),
        ),
      ],
      mainAxisAlignment: MainAxisAlignment.end,
    );
  }

  Future<bool> onBackPress() {
    {
//      Firestore.instance
//          .collection('users')
//          .document(id)
//          .updateData({'chattingWith': null});
      Navigator.pop(context);
    }

    return Future.value(false);
  }

  Widget buildTopBar() {
    return Builder(
      builder: (BuildContext context) => SafeArea(
        child: Column(
          children: <Widget>[
            Row(
              mainAxisAlignment: MainAxisAlignment.spaceBetween,
              children: <Widget>[
                Visibility(
                  visible: BlocProvider.of<ChatBloc>(context).isCustomer,
                  child: GestureDetector(
                    behavior: HitTestBehavior.translucent,
                    onTap: () => Navigator.of(context).pop(),
                    child: Padding(
                      padding: EdgeInsets.all(32),
                      child: Image.asset(
                        "packages/supchat_chat/assets/icons/top_bar_back.png",
                        width: 12,
                        height: 10,
                      ),
                    ),
                  ),
                ),
                BlocBuilder<ChatBloc, ChatState>(
                  builder: (BuildContext context, ChatState state) {
                    print('blocbuilder ${state.chatInfo}');
                    return Row(
                      crossAxisAlignment: CrossAxisAlignment.center,
                      children: <Widget>[
                        Container(
                          margin: EdgeInsets.only(right: 8),
                          height: 12,
                          width: 12,
                          decoration: BoxDecoration(
                              shape: BoxShape.circle,
                              color: state.chatInfo?.online?.online == 'online'
                                  ? Colors.green
                                  : Colors.grey),
                        ),
                        Container(
                          width: 120,
                          padding: EdgeInsets.symmetric(vertical: 24),
                          child: Text(
                            '${state.chatInfo?.name ?? state.chatInfo?.id ?? ''}',
                            maxLines: 1,
                            textAlign: TextAlign.center,
                            overflow: TextOverflow.ellipsis,
                            style: GoogleFonts.montserrat(
                                textStyle: TextStyle(
                                    fontSize: 15,
                                    color: Color(0xff0F2253),
                                    fontWeight: FontWeight.w500)),
                          ),
                        ),
                      ],
                    );
                  },
                ),
                Visibility(
                  visible: BlocProvider.of<ChatBloc>(context).isCustomer,
                  child: GestureDetector(
                    behavior: HitTestBehavior.translucent,
                    onTap: () {
                      widget.onDetailsClick?.call(context);
                    },
                    child: Padding(
                      padding: EdgeInsets.all(32),
                      child: Image.asset(
                        'packages/supchat_chat/assets/icons/top_bar_settings.png',
                        width: 14,
                        height: 4,
                      ),
                    ),
                  ),
                ),
              ],
            ),
            Divider(
              color: Color(0xffEDF1F9),
              height: 1,
              thickness: 1.0,
            ),
          ],
        ),
      ),
    );
  }

  Widget buildLoading() {
    return Positioned(
        child: Container(
      child: Center(
        child: CircularProgressIndicator(
            valueColor: AlwaysStoppedAnimation<Color>(themeColor)),
      ),
      color: Colors.white.withOpacity(0.8),
    ));
  }

  Widget buildInput(BuildContext context) {
    return SafeArea(
      top: false,
      child: Container(
        padding: EdgeInsets.symmetric(horizontal: 24, vertical: 18),
        child: Row(
          mainAxisAlignment: MainAxisAlignment.spaceBetween,
          children: <Widget>[
            // Button send image
            BlocBuilder<ChatBloc, ChatState>(
              builder: (BuildContext context, ChatState state) => new Container(
                child: GestureDetector(
                  behavior: HitTestBehavior.translucent,
                  onTap: () => getImage(context),
                  child: new Padding(
                    padding: EdgeInsets.all(8),
                    child: Image.asset(
                      'packages/${widget.chatConfig.isMobileLibrary ? 'supchat/supchat_chat/' : 'supchat_chat/'}assets/icons/send_attachment.png',
                      width: 18,
                      height: 18,
                      fit: BoxFit.fill,
                    ),
                  ),
                ),
              ),
            ),
            // Edit text
            Flexible(
              child: Container(
                margin: EdgeInsets.only(left: 16),
                child: Builder(
                  builder: (context) => TextField(
                    style: TextStyle(color: primaryColor, fontSize: 15.0),
                    controller: textEditingController,
                    decoration: InputDecoration.collapsed(
                      hintText: 'Type your message...',
                      hintStyle: TextStyle(color: greyColor),
                    ),
                    focusNode: focusNode,
                    onSubmitted: (text) {
                      onSendMessage(context, text, 0);
                    },
                  ),
                ),
              ),
            ),
            // Button send message
            Material(
              child: BlocBuilder<ChatBloc, ChatState>(
                builder: (BuildContext context, ChatState state) =>
                    new Container(
                  child: GestureDetector(
                    behavior: HitTestBehavior.translucent,
                    onTap: () =>
                        onSendMessage(context, textEditingController.text, 0),
                    child: Padding(
                        padding: const EdgeInsets.all(8.0),
                        child: Image.asset(
                          'packages/${widget.chatConfig.isMobileLibrary ? 'supchat/supchat_chat/' : 'supchat_chat/'}assets/icons/send_message.png',
                          width: 21,
                          height: 18,
                        )),
                  ),
                ),
              ),
              color: Colors.white,
            ),
          ],
        ),
        decoration: new BoxDecoration(
            border: new Border(
                top: new BorderSide(color: Color(0xffEDF1F9), width: 1)),
            color: Colors.white),
      ),
    );
  }

  Widget buildMessagesSection() {
    return BlocListener<ChatBloc, ChatState>(
        listenWhen: (_, state) =>
            state is NewMessageState || state is NewMessagesState,
        listener: (BuildContext context, ChatState state) {
          _messages = (state as MessagesHolder).messages;
          if (state is NewMessageState) {
            _chatMessagesGlobalKey.currentState
                ?.insertItem(0, duration: Duration(milliseconds: 200));
          }
        },
        child: BlocBuilder<ChatBloc, ChatState>(
          buildWhen: (prevState, state) =>
              !(state is NewMessageState) && prevState != state,
          builder: (BuildContext context, state) => Expanded(
            child: Builder(
              builder: (context) {
                if (state is ChatLoaded) {
                  return buildMessagesList(state.messages);
                } else if (state is ImageLoading) {
                  return Stack(children: <Widget>[
                    buildMessagesList(state.messages),
                    buildLoading()
                  ]);
                } else {
                  return Center(
                      child: CircularProgressIndicator(
                          valueColor:
                              AlwaysStoppedAnimation<Color>(themeColor)));
                }
              },
            ),
          ),
        ));
  }

  GlobalKey<AnimatedListState> _chatMessagesGlobalKey =
      GlobalKey<AnimatedListState>();

  List<Message> _messages = [];

  Widget buildMessagesList(List<Message> messages) {
    _messages = messages;
    return AnimatedList(
      key: _chatMessagesGlobalKey,
      padding: EdgeInsets.symmetric(horizontal: 32.0, vertical: 18),
      itemBuilder: (context, index, animation) {
        return index >= _messages.length
            ? Container()
            : SlideTransition(
                position: Tween<Offset>(
                  begin: Offset(0.1, 0),
                  end: Offset.zero,
                ).animate(animation),
                child: SizeTransition(
                    sizeFactor: CurvedAnimation(
                        parent: animation, curve: Interval(0.0, 1.0)),
                    axisAlignment: 0.0,
                    child: buildItem(index, _messages[index])));
      },
      reverse: true,
      initialItemCount: _messages.length,
      controller: listScrollController,
    );
  }
}

extension on DateTime {
  String formatTime() {
    final dateFormat = DateFormat('HH:MM');
    return dateFormat.format(this);
  }
}

final DateFormat _dateFormat = DateFormat('dd MMMM');

extension on Message {
  String headerText(Message prevMessage) {
    if (time.day == prevMessage?.time?.day) return null;
    return _dateFormat.format(time);
  }
}

extension on List<Message> {
  Message safeGet(int index) {
    if (length <= index || index < 0) return null;
    return this[index];
  }
}
